part of firebase_model_notifier;

/// Class for executing Firestore Transaction.
class FirestoreTransaction {
  const FirestoreTransaction._();

  /// Outputs the builder to be written by the transaction.
  ///
  /// Basically, it writes and deletes data for [collectionPath].
  ///
  /// You can add the corresponding element by specifying [linkedCollectionPath].
  ///
  /// If [enableCounter] is set to `true`,
  /// the number of elements will be counted and recorded in the document of the level above each collection path.
  ///
  /// You can generate a key to store the number of elements in a document by specifying [counterBuilder] or [linkedCounterBuilder].
  static FirestoreDatabaseTransactionBuilder transaction({
    required String collectionPath,
    bool enableCounter = true,
    String counterSuffix = "Count",
    String Function(String path)? counterBuilder,
    String? linkedCollectionPath,
    String Function(String linkPath)? linkedCounterBuilder,
    List<CounterUpdaterInterval> counterIntervals = const [],
    DynamicMap? additionalData,
    DynamicMap? linkedAdditionalData,
  }) {
    return FirestoreDatabaseTransactionBuilder._(
      collectionPath: collectionPath,
      enableCounter: enableCounter,
      counterSuffix: counterSuffix,
      counterBuilder: counterBuilder,
      linkedCollectionPath: linkedCollectionPath,
      linkedCounterBuilder: linkedCounterBuilder,
      counterIntervals: counterIntervals,
      additionalData: additionalData,
      linkedAdditionalData: linkedAdditionalData,
    );
  }

  /// Create a code of length [length] randomly for id.
  ///
  /// Characters that are difficult to understand are omitted.
  ///
  /// If the data of [key] in [path] contains the generated random value,
  /// the random value is generated again.
  static Future<String> generateCode({
    required String path,
    required String key,
    int length = 6,
    String charSet = "23456789abcdefghjkmnpqrstuvwxy",
  }) async {
    await FirebaseCore.initialize();
    List<String> generated = [];
    do {
      // await Future.delayed(Duration(milliseconds: Random().nextInt(100)));
      generated = List.generate(
          10, (index) => katana.generateCode(length, charSet: charSet));
      final snapshot = await FirebaseFirestore.instance
          .collection(path)
          .where(key, whereIn: generated)
          .get();
      for (final doc in snapshot.docs) {
        if (!doc.exists) {
          continue;
        }
        final map = doc.data();
        if (map.isEmpty || !map.containsKey(key)) {
          continue;
        }
        generated.remove(map.get(key, ""));
      }
    } while (generated.isEmpty);
    return generated.first;
  }
}

/// Generates a transaction builder.
class FirestoreDatabaseTransactionBuilder extends DatabaseTransactionBuilder {
  FirestoreDatabaseTransactionBuilder._({
    required this.collectionPath,
    this.counterBuilder,
    this.enableCounter = true,
    this.counterSuffix = "Count",
    this.linkedCollectionPath,
    this.linkedCounterBuilder,
    this.counterIntervals = const [],
    this.additionalData,
    this.linkedAdditionalData,
  })  : assert(
            !(collectionPath.splitLength() <= 0 ||
                collectionPath.splitLength() % 2 != 1),
            "The collection path hierarchy must be an odd number."),
        assert(
            linkedCollectionPath.isEmpty ||
                !(linkedCollectionPath!.splitLength() <= 0 ||
                    linkedCollectionPath.splitLength() % 2 != 1),
            "The link collection path hierarchy must be an odd number.");

  /// The counter is up.
  final bool enableCounter;

  /// Suffix to generate the counter key.
  final String counterSuffix;

  /// A collection that counts the number of elements.
  final String collectionPath;

  /// Builder to generate the counter key.
  final String Function(String path)? counterBuilder;

  /// A linked collection that counts the number of elements.
  final String? linkedCollectionPath;

  /// Builder to generate the counter key for linked collection.
  final String Function(String linkPath)? linkedCounterBuilder;

  /// List of counters with valid expiration dates.
  final List<CounterUpdaterInterval> counterIntervals;

  /// Additional data in the counter document.
  final DynamicMap? additionalData;

  /// Additional data for linked counter documents.
  final DynamicMap? linkedAdditionalData;

  String? _buildCounterPath(String? path) {
    if (path.isEmpty) {
      return null;
    }
    return "$path$counterSuffix";
  }

  DynamicMap _buildCounterUpdate(String key, num value) {
    if (!enableCounter) {
      return {};
    }
    final now = DateTime.now();
    final map = {key: FieldValue.increment(value)};
    for (final interval in counterIntervals) {
      switch (interval) {
        case CounterUpdaterInterval.daily:
          map[dailyKey(key, now)] = FieldValue.increment(value);
          for (var i = 0; i < 30; i++) {
            map[dailyKey(
                    key, DateTime(now.year, now.month, now.day - 60 + i))] =
                FieldValue.delete();
          }
          break;
        case CounterUpdaterInterval.monthly:
          map[monthlyKey(key, now)] = FieldValue.increment(value);
          for (var i = 0; i < 12; i++) {
            map[monthlyKey(key, DateTime(now.year, now.month - 24 + i))] =
                FieldValue.delete();
          }
          break;
        case CounterUpdaterInterval.yearly:
          map[yearlyKey(key, now)] = FieldValue.increment(value);
          for (var i = 0; i < 5; i++) {
            map[yearlyKey(key, DateTime(now.year, now.month - 10 + i))] =
                FieldValue.delete();
          }
          break;
        case CounterUpdaterInterval.weekly:
          map[weeklyKey(key, now)] = FieldValue.increment(value);
          for (var i = 0; i < 4; i++) {
            map[weeklyKey(key,
                    DateTime(now.year, now.month, now.day - ((8 - i) * 7)))] =
                FieldValue.delete();
          }
          break;
      }
    }
    return map;
  }

  /// Add/Update an element of [id] to [collectionPath].
  ///
  /// If [linkedCollectionPath] is specified, the element of [linkId] is added.
  ///
  /// The contents of [data] will be added to the [collectionPath]/[id] path.
  ///
  /// The contents of [linkedData] will be added to the [linkedCollectionPath]/[linkId] path.
  @override
  Future<void> save(
    String id, {
    String? linkId,
    DynamicMap? data,
    DynamicMap? linkData,
  }) async {
    assert(
        linkId.isEmpty ||
            (linkedCollectionPath.isNotEmpty && linkId.isNotEmpty),
        "When [linkId] is specified, [linkPath] must be specified.");
    await FirebaseCore.initialize();
    // await Future.delayed(Duration(milliseconds: Random().nextInt(100)));
    final firestore = FirebaseFirestore.instance;
    final docPath = counterBuilder?.call(collectionPath) ??
        _buildCounterPath(collectionPath) ??
        "";
    final linkDocPath = linkedCollectionPath.isEmpty
        ? null
        : (linkedCounterBuilder?.call(linkedCollectionPath!) ??
            _buildCounterPath(linkedCollectionPath!) ??
            "");
    await firestore.runTransaction((transaction) async {
      try {
        DocumentSnapshot<Map<String, dynamic>>? doc;
        DocumentSnapshot<Map<String, dynamic>>? linkDoc;
        await Future.wait([
          transaction
              .get(firestore.doc("$collectionPath/$id"))
              .then((value) => doc = value),
          if (!(linkId.isEmpty || linkedCollectionPath.isEmpty))
            transaction
                .get(firestore.doc("$linkedCollectionPath/$linkId"))
                .then((value) => linkDoc = value)
        ]);
        if (doc != null && !doc!.exists) {
          transaction.set(
            doc!.reference,
            {
              if (data != null) ...data,
              Const.uid: id,
              Const.time: FieldValue.serverTimestamp(),
            },
            SetOptions(merge: true),
          );
          if (docPath.isNotEmpty) {
            final key = docPath.split("/").last;
            final path = docPath.replaceFirst(RegExp("/$key\$"), "");
            final countDoc = {};
            additionalData?.forEach((key, value) {
              if (value == null || key.isEmpty) {
                return;
              }
              countDoc[key] = value;
            });
            countDoc.addAll(_buildCounterUpdate(key, 1));
            transaction.set(
              firestore.doc(path),
              countDoc,
              SetOptions(merge: true),
            );
          }
        }
        if (linkDoc != null && !linkDoc!.exists) {
          transaction.set(
            linkDoc!.reference,
            {
              if (linkData != null) ...linkData,
              Const.uid: linkId,
              Const.time: FieldValue.serverTimestamp(),
            },
            SetOptions(merge: true),
          );
          if (linkDocPath.isNotEmpty) {
            final key = linkDocPath!.split("/").last;
            final path = linkDocPath.replaceFirst(RegExp("/$key\$"), "");
            final countDoc = {};
            linkedAdditionalData?.forEach((key, value) {
              if (value == null || key.isEmpty) {
                return;
              }
              countDoc[key] = value;
            });
            countDoc.addAll(_buildCounterUpdate(key, 1));
            transaction.set(
              firestore.doc(path),
              countDoc,
              SetOptions(merge: true),
            );
          }
        }
      } catch (e) {
        print(e.toString());
        rethrow;
      }
    });
  }

  /// Remove an element of [id] to [collectionPath].
  ///
  /// If [linkedCollectionPath] is specified, the element of [linkId] is removed.
  @override
  Future<void> delete(String id, {String? linkId}) async {
    assert(
        linkId.isEmpty ||
            (linkedCollectionPath.isNotEmpty && linkId.isNotEmpty),
        "When [linkId] is specified, [linkPath] must be specified.");
    await FirebaseCore.initialize();
    final firestore = FirebaseFirestore.instance;
    final docPath = counterBuilder?.call(collectionPath) ??
        _buildCounterPath(collectionPath) ??
        "";
    final linkDocPath = linkedCollectionPath.isEmpty
        ? null
        : (linkedCounterBuilder?.call(linkedCollectionPath!) ??
            _buildCounterPath(linkedCollectionPath!) ??
            "");
    await firestore.runTransaction((transaction) async {
      try {
        DocumentSnapshot<Map<String, dynamic>>? doc;
        DocumentSnapshot<Map<String, dynamic>>? linkDoc;
        await Future.wait([
          transaction
              .get(firestore.doc("$collectionPath/$id"))
              .then((value) => doc = value),
          if (!(linkId.isEmpty || linkedCollectionPath.isEmpty))
            transaction
                .get(firestore.doc("$linkedCollectionPath/$linkId"))
                .then((value) => linkDoc = value)
        ]);
        if (doc != null && doc!.exists) {
          transaction.delete(doc!.reference);
          if (docPath.isNotEmpty) {
            final key = docPath.split("/").last;
            final path = docPath.replaceFirst(RegExp("/$key\$"), "");
            final countDoc = {};
            additionalData?.forEach((key, value) {
              if (value == null || key.isEmpty) {
                return;
              }
              countDoc[key] = value;
            });
            countDoc.addAll(_buildCounterUpdate(key, -1));
            transaction.set(
              firestore.doc(path),
              countDoc,
              SetOptions(merge: true),
            );
          }
        }
        if (linkDoc != null && linkDoc!.exists) {
          transaction.delete(linkDoc!.reference);
          if (linkDocPath.isNotEmpty) {
            final key = linkDocPath!.split("/").last;
            final path = linkDocPath.replaceFirst(RegExp("/$key\$"), "");
            final countDoc = {};
            linkedAdditionalData?.forEach((key, value) {
              if (value == null || key.isEmpty) {
                return;
              }
              countDoc[key] = value;
            });
            countDoc.addAll(_buildCounterUpdate(key, -1));
            transaction.set(
              firestore.doc(path),
              countDoc,
              SetOptions(merge: true),
            );
          }
        }
      } catch (e) {
        print(e.toString());
        rethrow;
      }
    });
  }
}
